%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2001/10/22 23:08:07 $ by $Author: afranke $
%%%   $Revision: 1.8 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% XmlScanner.oz
%%
%% tokenizes DTDs and XML documents
%%

functor
   
import
%   Boot_Debug at 'x-oz://boot/Debug'
   Compiler(virtualStringToValue)
   Error(registerFormatter)
   Open(file)
   System(/*show*/ showError /*showInfo printInfo*/)
   
   X('class': FIFO) at 'FIFO.ozf'
   XVS              at 'XVS.ozf'
   
export
   'class':  Scanner

   dtd:      DTD      % rec(scan:       fun{$ S BaseUrl}
                      %     scanString: fun{$ S}
                      %     scanVS:     fun{$ VS}
                      %     scanFile:   fun{$ FileName}
                      %     scanUrl:    fun{$ Url}
                      %    )
   
   document: Document % rec(scan:       fun{$ S BaseUrl}
                      %     scanString: fun{$ S}
                      %     scanVS:     fun{$ VS}
                      %     scanFile:   fun{$ FileName}
                      %     scanUrl:    fun{$ Url}
                      %    )
   
define
   
   proc {Skip} skip end

   proc {ShowDebug X}
      skip
      %%{System.showInfo {Value.toVirtualString X 10 20}}
   end
   
   VS2S = VirtualString.toString
   VS2A = VirtualString.toAtom
   
   %%------------------------------------------------------------
   %% Character Classes 
   %%------------------------------------------------------------
   
   CharClass = {NewArray 0 255 unit}
   local
      proc {DefChar C CType}
         {Put CharClass C CType}
      end
      proc {DefInterval C1 C2 CType}
         {For C1 C2 1 proc {$ C} {DefChar C CType} end}
      end
      proc {Define Defs CType}
         {ForAll Defs
          proc {$ Def}
             case Def of C1#C2 then {DefInterval C1 C2 CType}
             [] C then {DefChar C CType} end
          end}
      end
      DefSpace         = [&\x20 &\x09 &\x0D &\x0A]
      DefBaseChar      = [&\x41#&\x5A % A-Z
                          &\x61#&\x7A % a-z
                          &\xC0#&\xD6
                          &\xD8#&\xF6
                          &\xF8#&\xFF]
      DefIdeographic   = nil
      DefCombiningChar = nil
      DefDigit         = [&\x30#&\x39] % 0-9
      DefExtender      = [&\xB7]
   in
      {ForAll [DefSpace        #space
               DefBaseChar     #basechar
               DefIdeographic  #ideographic
               DefCombiningChar#combiningchar
               DefDigit        #digit
               DefExtender     #extender
              ]
       proc {$ Defs#CType} {Define Defs CType} end}
   end

   fun {IsLetterChar C}
      T={Get CharClass C} in
      T==basechar orelse T==ideographic
   end
   /*
   fun {IsDigitChar C}
      {Get CharClass C}==digit
   end
   */
   fun {IsSpaceChar C}
      {Get CharClass C}==space
   end

   %% general character tests
   fun {MakeCharTest Members}
      Chars = {BitArray.new 0 255}
      proc {DefC C} {BitArray.set Chars C} end
      proc {DefCs C1 C2} {For C1 C2 1 DefC} end 
      proc {DefX X}
         case X of C1#C2 then {DefCs C1 C2} else {DefC X} end
      end
   in
      {ForAll Members DefX}
      fun {$ C} {BitArray.test Chars C} end
   end
   
% [4] NameChar ::=  Letter | Digit |'.'|'-'|'_'|':'| CombiningChar | Extender
   
   fun {IsNameChar C}
      T={Get CharClass C} in
      {Member T [basechar ideographic digit combiningchar extender]} orelse
      {Member C [&_ &. &: &-]}
   end

% [13] PubidChar ::= #x20 | #xD | #xA | [a-zA-Z0-9] | [-'()+,./:=?;!*#@$_%]

   IsPubidChar = {MakeCharTest
                  &\x20|&\x0D|&\x0A|&a#&z|&A#&Z|&0#&9|"-'()+,./:=?;!*#@$_%"}
   
% [26] VersionNum ::=  ([a-zA-Z0-9_.:] | '-')+

   IsVNChar = {MakeCharTest &a#&z|&A#&Z|&0#&9|"_.:-"}

% [81] EncName ::=  [A-Za-z] ([A-Za-z0-9._] | '-')*

   IsAZaz        = {MakeCharTest [&A#&Z &a#&z]}
   IsEncNameChar = {MakeCharTest  &A#&Z|&a#&z|&0#&9|"._-"}   


   %%------------------------------------------------------------
   %% Support for Character & Entity References 
   %%------------------------------------------------------------

   fun {DropZeroes Dec}
      case Dec of nil then "0"
      [] &0|R then {DropZeroes R}
      else Dec end
   end

   local
      proc {Split S Spec ?S1 ?S2}
         case S of nil then S1=nil S2=nil
         [] C|Sr then
            if {HasFeature Spec C} then S1=nil S2=S 
            else S1=C|{Split Sr Spec $ ?S2} end
         end
      end
   in
      fun {Parse S Spec}
         case S of nil then nil
         else
            S1 S2 R in
            {Split S Spec ?S1 ?S2}
            R = {ParseRef S2 Spec}
            case S1 of nil then R
            else S1#R end
         end
      end
      fun {ParseRef S Spec}
         case S of nil then nil
         [] C|Sr then
            S1 S2 in
            {String.token Sr &; ?S1 ?S2}
            {MakeRef Spec.C S1}#{Parse S2 Spec}
         end
      end
      fun {MakeRef L S} L({StringToAtom S}) end
   end

   local
      fun {Dirname Path}
         {Reverse {String.token {Reverse Path} &/ _}}#"/"
      end
      fun {IsAbsolute Url}
         {List.isPrefix "http:" Url} orelse
         {List.isPrefix "file:" Url} orelse
         {List.isPrefix "/" Url}
      end
   in
      fun {MkUrl BaseUrl SystemID}
         if {IsAbsolute SystemID}
         then SystemID
         elseif {IsAbsolute BaseUrl}
         then {Dirname BaseUrl}#SystemID
         else SystemID
         end
      end
   end
   proc {LoadUrlWithTail Url Tail ?S}
      F = {New Open.file init(url:Url)} in
      {F read(list:?S size:all tail:Tail)}
      {F close}
   end
   
   EntityValueSpec = unit(&%:'%' &&:'\&')
   AttValueSpec    = unit(&&:'\&')
   
   %%
   %% Refs Registry
   %%
  
   class Refs
      feat
         GEs
         PEs
      meth init
         self.GEs = {NewDictionary}
         self.PEs = {NewDictionary}
      end
      meth enterPE(Key Def) = M
         {ShowDebug M}
         {Dictionary.put self.PEs Key Def}
      end
      meth enterGE(Key Def) = M
         {ShowDebug M}
         {Dictionary.put self.GEs Key Def}
      end
      meth lookupPE(Key $) = M
         {ShowDebug M}
         case {Dictionary.condGet self.PEs Key unit} of unit then
            Scanner, reportErr('undefined parameter entity' hints:[name#Key])
            dummy
         [] Def then Def
         end
      end
      meth lookupGE(Key $) = M
         {ShowDebug M}
         case {Dictionary.condGet self.GEs Key unit} of unit then
            Scanner, reportErr('undefined general entity' hints:[name#Key])
            dummy
         [] Def then Def
         end
      end
      meth replacePE(Key $) = M
         {ShowDebug M}
         Def = Refs, lookupPE(Key $) in
         case Def of 'VAL'(Val) then
            {XVS.map Val fun {$ Ref}
                            case Ref of '%'(N) then Refs, replacePE(N $)
                            else Ref %% really keep GEs?
                            end
                         end}
         else
            Scanner, reportErr('externalPE'(Key Def) quiet:true) dummy
         end      
      end
      meth replaceGE(Key $) = M
         {ShowDebug M}
         unimplemented
      end
   end
   
   %% {VirtualString.toString {D lookup(test $)}}
   %% {Parse "hallo%test;foo&bla;bar" unit(&%:'%' &&:'\&'})
   
   %%------------------------------------------------------------
   %% Summary of Handling CharRefs, EntityRefs & PERefs
   %%------------------------------------------------------------
   
   %% while processing DTD:
   %% -- recognize PEs and process replacement text instead (recursively!)
   
   %% while processing entity values:
   %% -- recognize PCs      and replace;
   %% -- recognize CharRefs and replace;
   %% -- recognize Refs     and bypass;

   %% while processing content or attribute values:
   %% -- recognize CharRefs and replace;
   %% -- recognize Refs     and replace;

   %% these features are not fully implemented yet
   
   %%------------------------------------------------------------
   %% The XML Scanner Class
   %%------------------------------------------------------------

   BufferInfo = [url s line col ghost substPE] 
   
   class Scanner from Refs

      feat
         onlyOnce
         
      attr
         bufferStack: nil
         
         url:''
         s: nil
     
         line: 1
         col:  1
         
         ghost: 0 %% number of chars appended (entity replacement) 

         substPE: false

         includeExternalPEs: true
         
         tokens: unit
 
      meth init(S url:Url<='')
         self.onlyOnce = {NewName}

         Scanner, InitBuffer(S url:Url)
         
         tokens <- {New FIFO init}
         
         Refs, init
      end
      
      %% buffer stack
      meth InitBuffer(S url:Url)         
         url <- {VirtualString.toString Url}
         
         Scanner, CheckString(S)
         s <- S
         
         line <- 1
         col  <- 1
         
         ghost <- 0
      end
      meth PushBuffer
         B = {MakeRecord buffer BufferInfo} in
         {ForAll BufferInfo proc {$ F} B.F=@F end}
         bufferStack <- B|@bufferStack
      end
      meth PopBuffer
         B|Rest = @bufferStack in
         bufferStack <- Rest
         {ForAll BufferInfo proc {$ F} F<-B.F end}
      end
      
      meth CheckString(S)
         case S
         of _|_ then skip
         [] nil then skip
         else raise scanner('Can only read from strings.') end
         end
      end

      %% switch for automatic parameter entity substitution
      meth substPEOn
         substPE<-true
         if @ghost == 0 then
            Scanner, OptSubstPE
         end
      end
      meth substPEOff
         substPE<-false
      end
      
      %% token stream
      meth putToken(T Val)
         {ShowDebug T#Val}
         {@tokens put(T#Val)}
      end
      meth putToken1(T)
         Scanner, putToken(T unit)
      end
      meth getToken(?T ?Val)
         {@tokens get(?T ?Val)}
      end
      meth getTokens(?L)
         {@tokens getAll(?L)}
      end

      %%
      %% basic operations (getC, ungetC)
      %%
      meth readChar($)
         case @s of nil then false 
         [] C|Sr then
            if C == &\n then line<-@line+1 col<-0 
            else col<-@col+1 end
            s<-Sr C
         end
      end
      meth getRest($)
         @s
      end

      %%------------------------------------------------------------
      %% BEGIN OF XML-SPECIFIC METHODS 
      %%------------------------------------------------------------

      %%
      %% error reporting
      %%
      meth ShowRest
         {System.showError "*** next 80 characters ***"}
         if {List.drop @s 80}\=nil then
            {System.showError {List.take @s 80}}
         else
            {System.showError @s}
         end
         {System.showError "*** end ***"}
      end
      meth reportErr(MsgA hints:Hints<=nil quiet:Quiet<=false)
         if Quiet==false then
            Scanner, ShowRest
         end
         Pos = Scanner, getPos($) in
         %% {self reportError(Pos MsgA InfoList)}
         {Exception.raiseError scanner(MsgA hints:Hints pos:Pos)}
      end
      meth getPos($)
         pos('' @line @col)
      end
      
      %%
      %% character basics
      %%
      meth atEnd($)
         @s==nil
      end
      meth getC(?Char)
	 case @s of nil then
	    Scanner, reportErr('unexpected end of file')
         [] C|Sr then
            Char=C
            s<-Sr
            Scanner, incPos(C)
         end
         if @substPE andthen @ghost == 0 then
            Scanner, OptSubstPE
         end
      end
      meth OptSubstPE
         if Scanner, IsPEReference($) then
            N ReplacementVS in
            Scanner, PEReferenceA(?N)
            try
               Scanner, replacePE(N ?ReplacementVS)
            catch error(scanner(externalPE(N _/*Def*/)...)...) then
               %% For external parameter entities, restore the original
               %% reference; it will be handled explicitly in meth DtdItems
               ReplacementVS="%"#N#";"
            end
            Scanner, addPrefix(ReplacementVS)
         end
      end
      meth topC($)
	 case @s of nil then false
	 [] C|_ then C end
      end
      meth incPos(C)
         if @ghost > 0 then
            ghost <- @ghost - 1
         elseif C==&\n then
            line <- @line + 1
            col <- 0
         else
            col <- @col + 1
         end
      end
      meth addPrefix(VS)
         S = {VS2S VS} in
         s<-{Append S @s}
         ghost <- @ghost + {Length S}
      end
      meth skipC
         Scanner, getC(_)
      end
      meth Chr(ExpectedC)
         C = Scanner, getC($) in
         if C\=ExpectedC then
            Scanner, reportErr({VS2A '"'#[ExpectedC]#'" expected'})
         end
      end
      
      %%
      %% string basics
      %%
      meth IsPrefix(Prefix $)
         {List.isPrefix Prefix @s}
      end
      meth skipN(N)
         if N==0 then skip
         else
            Scanner, skipC
            Scanner, skipN(N-1)
         end
      end
      meth Str(ExpectedS)
         if Scanner, IsPrefix(ExpectedS $) then
            Scanner, skipN({Length ExpectedS})
         else
            Scanner, reportErr({VS2A '"'#ExpectedS#'" expected'})
         end
      end
      
      %%
      %% read characters sequences
      %%
      meth ReadUntilC(StopC $) % inclusive
         C = Scanner, getC($) in
         if C==StopC then nil
         else
            C | Scanner, ReadUntilC(StopC $)
         end
      end
      meth ReadUntilCs(StopCs $) % exclusive
         C = Scanner, topC($) in
         if {Member C StopCs} then nil
         else
            Scanner, skipC
            C | Scanner, ReadUntilCs(StopCs $)
         end
      end
      meth ReadUntilM(StopCond $) % exclusive
         if Scanner, StopCond($) then nil
         else
            C = Scanner, getC($) in
            C | Scanner, ReadUntilM(StopCond $)
         end
      end
      meth ReadUntilS(StopS $) % inclusive
         S1 | Sr = StopS in
         Scanner, DoReadUntilS(S1 Sr $)
      end
      meth DoReadUntilS(S1 Sr $)
         C = Scanner, getC($) in
         if C==S1 andthen Scanner, IsPrefix(Sr $) then
            Scanner, skipN({Length Sr}) nil
         else
            C | Scanner, DoReadUntilS(S1 Sr $)
         end
      end
      
%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%  Extensible Markup Language (XML) 1.0
%%%  W3C Recommendation 10-February-1998
%%%  http://www.w3.org/TR/1998/REC-xml-19980210
%%%
%%%  Grammar Rules 
%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

      %%
      %% Note: To make things easier, the implementation has moved all
      %%       (evtl. optional) leading spaces from "inner" (sub-)rules
      %%       to the calling "outer" rules. 
      %%       Similarly, if two alternatives were beginning with the same
      %%       pattern, this may have been moved to the calling "outer" rules.
      %%
      
%%%
%%% 2.1 Well-Formed XML Documents
%%%

% [1] document ::=  prolog element Misc*

      meth document
         %% hack
         includeExternalPEs <- false
         Scanner, Prolog
         Scanner, Element
         Scanner, OptMiscs
         if Scanner, atEnd($) then
            Scanner, putToken1('EOF')
         else
            Scanner, reportErr('end-of-file expected')
         end
      end
      
% %%
% %% 2.3 Common Syntactic Constructs
% %%

% %% White Space

% [3] S ::=  (#x20 | #x9 | #xD | #xA)+

      meth IsSpace($)
	 C = Scanner, topC($) in
         C\=false andthen {IsSpaceChar C}
      end
      meth Spaces
         if Scanner, IsSpace($) then
            Scanner, skipC
            Scanner, OptSpaces
         else
            {self reportErr('space expected')}
         end
      end
      meth OptSpaces
         if Scanner, IsSpace($) == false then skip
         else
            Scanner, skipC
            Scanner, OptSpaces
         end
      end
      
% %% Names and Tokens

% [4] NameChar ::=  Letter | Digit | '.' | '-' | '_' | ':' | CombiningChar | Extender      
% [5]     Name ::=  (Letter | '_' | ':') (NameChar)*
% [6]    Names ::=  Name (S Name)*
% [7]  Nmtoken ::=  (NameChar)+
% [8] Nmtokens ::=  Nmtoken (S Nmtoken)*

      meth ReadNameChars($)
         C = Scanner, topC($) in
         if {IsNameChar C}==false then nil
         else
            Scanner, skipC
            C | Scanner, ReadNameChars($)
         end
      end
      meth IsName($)
         C = Scanner, topC($) in
         {IsLetterChar C} orelse C==&_ orelse C==&:
      end
      meth Name($)
         if Scanner, IsName($) then
            C = Scanner, getC($) in
            C | Scanner, ReadNameChars($)
         else
            Scanner, reportErr('name expected') dummy
         end
      end
      meth NameA($)
         N = Scanner, Name($) in
         {StringToAtom N}
      end
      meth IsNmtoken($)
         C = Scanner, topC($) in
         {IsNameChar C}
      end
      meth Nmtoken($)
         if Scanner, IsNmtoken($) then
            C = Scanner, getC($) in
            C | Scanner, ReadNameChars($)
         else
            Scanner, reportErr('name token expected') dummy
         end
      end
      meth NmtokenA($)
         N = Scanner, Nmtoken($) in
         {StringToAtom N}
      end
   
% %% Literals

% [9]    EntityValue ::= '"' ([^%&"] | PEReference | Reference)* '"' 
%                         |  "'" ([^%&'] | PEReference | Reference)* "'"
% [10]      AttValue ::= '"' ([^<&"] | Reference)* '"' 
%                         |  "'" ([^<&'] | Reference)* "'"
% [11] SystemLiteral ::=  ('"' [^"]* '"') | ("'" [^']* "'") 
% [12]  PubidLiteral ::= '"' PubidChar* '"' | "'" (PubidChar - "'")* "'"
% [13]     PubidChar ::= #x20 | #xD | #xA | [a-zA-Z0-9] | [-'()+,./:=?;!*#@$_%]

      meth IsLiteral($)
         C = Scanner, topC($) in
         C==&" orelse C==&'
      end
      meth Literal(?S)
         SubstPE=@substPE
         substPE<-false
         QuotC = Scanner, getC($) in
         Scanner, ReadUntilC(QuotC ?S)
         substPE<-SubstPE
      end

      meth IsEntityValue($)   Scanner, IsLiteral($) end
      meth IsAttValue($)      Scanner, IsLiteral($) end
      meth IsSystemLiteral($) Scanner, IsLiteral($) end
      meth IsPubidLiteral($)  Scanner, IsLiteral($) end
      
      meth EntityValue(?Val)
         Lit = Scanner, Literal($) in
         Val = {Parse Lit EntityValueSpec}
      end
      meth AttValue($)
         Lit = Scanner, Literal($) in
         {Parse Lit AttValueSpec}
      end
      meth SystemLiteral(?Lit)
         Scanner, Literal(?Lit)
      end
      meth PubidLiteral(?Lit)
         Scanner, Literal(?Lit)
         if {All Lit IsPubidChar}==false then
            Scanner, reportErr('illegal character in "PUBLIC" identifier')
         end
      end
      
% %%
% %% 2.4 Character Data and Markup
% %%

% [14] CharData ::=  [^<&]* - ([^<&]* ']]>' [^<&]*)
      
      meth CharData %% ]]> abfangen!
         S = Scanner, ReadUntilCs([&< && &]] $) in
         Scanner, putToken('characters' S)
         if Scanner, topC($) == &] andthen
            Scanner, IsPrefix("]]>" $) == false then
            Scanner, Chr(&])
            Scanner, putToken('characters' "]")
            Scanner, CharData
         end
      end
      
% %%
% %% 2.5 Comments
% %%

% [15] Comment ::=  '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'

      meth IsComment($)
         Scanner, IsPrefix("<!--" $)
      end
      meth Comment
         SubstPE=@substPE in
         substPE<-false
         Scanner, Str("<!--")
         Scanner, ReadUntilS("-->" _)
         substPE<-SubstPE
      end
      
% %%
% %% 2.6 Processing Instructions
% %%

% [16]       PI ::=  '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
% [17] PITarget ::=  Name - (('X' | 'x') ('M' | 'm') ('L' | 'l'))

      meth IsPI($)
	 Scanner, IsPrefix("<?" $) 
      end
      meth PI
         TargetS CharsS PIVal in
         Scanner, Str("<?")
         Scanner, Name(?TargetS)
         case {Map TargetS Char.toLower} of &x|&m|&l|R then
            if {List.isPrefix "-stylesheet" R}==false then
               Scanner, reportErr('pi target must not start with xml')
            end
         else skip end
         if Scanner, IsPrefix("?>" $) then
            CharsS = nil
            Scanner, Str("?>")
         else
            Scanner, Spaces
            Scanner, ReadUntilS("?>" ?CharsS)
         end
         PIVal = 'PI'({ByteString.make TargetS}
                      {ByteString.make CharsS})
         Scanner, putToken('PI' PIVal)
      end
      
% %%
% %% 2.7 CDATA Sections
% %%

% [18]  CDSect ::=  CDStart CData CDEnd
% [19] CDStart ::=  '<![CDATA['
% [20]   CData ::=  (Char* - (Char* ']]>' Char*)) 
% [21]   CDEnd ::=  ']]>'

      meth IsCDSect($)
         Scanner, IsPrefix("<![CDATA[" $)
      end
      meth CDSect
         CDataS in
         Scanner, Str("<![CDATA[") 
         Scanner, ReadUntilS("]]>" ?CDataS)
         Scanner, putToken('CDATA' {ByteString.make CDataS})
      end
      
% %%
% %% 2.8 Prolog and Document Type Declaration
% %%

% %% Prolog

% [22]      prolog ::=  XMLDecl? Misc* (doctypedecl Misc*)?
% [23]     XMLDecl ::=  '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
% [24] VersionInfo ::=  S 'version' Eq (' VersionNum ' | " VersionNum ")
% [25]          Eq ::=  S? '=' S?
% [26]  VersionNum ::=  ([a-zA-Z0-9_.:] | '-')+
% [27]        Misc ::=  Comment | PI |  S

      meth Prolog
         if Scanner, IsXMLDecl($) then
            Scanner, XMLDecl
         end
         Scanner, OptMiscs
         if Scanner, IsDoctypedecl($) then
            Scanner, Doctypedecl
            Scanner, OptMiscs
         end
      end
      meth IsXMLDecl($)
         Scanner, IsPrefix("<?xml" $)
      end
      meth XMLDecl
         Ver Enc SD in
         Scanner, Str("<?xml")
         Scanner, Spaces
         if Scanner, IsVersionInfo($) then
            Scanner, VersionInfo(?Ver)
         else
            Scanner, reportErr('version info expected')
         end
         Scanner, OptSpaces
         if Scanner, IsEncodingDecl($) then
            Scanner, EncodingDecl(?Enc)
         else
            Enc=unit
         end
         Scanner, OptSpaces
         if Scanner, IsSDDecl($) then
            Scanner, SDDecl(?SD)
         else
            SD=unit
         end
         Scanner, OptSpaces
         Scanner, Str("?>")
      end
      meth IsVersionInfo($)
         Scanner, IsPrefix("version" $)
      end
      meth VersionInfo(?VN)
         Quot in
         Scanner, Str("version")
         Scanner, Eq
         Scanner, getC(?Quot)
         if Quot\=&' andthen Quot\=&" then
            Scanner, reportErr('Quotes expected')
         else
            Scanner, VersionNum(?VN)
            Scanner, Chr(Quot)
         end
      end
      meth Eq
         Scanner, OptSpaces
         Scanner, Chr(&=)
         Scanner, OptSpaces
      end
      meth VersionNum($)
         C = Scanner, topC($) in
         if {IsVNChar C} then
            Scanner, skipC
            C | Scanner, ReadVNChars($)
         else
            Scanner, reportErr('VersionNum character expected') dummy
         end
      end
      meth ReadVNChars($)
         C = Scanner, topC($) in
         if {IsVNChar C}==false then nil
         else
            Scanner, skipC
            C | Scanner, ReadVNChars($)
         end
      end
      meth OptMiscs
         Scanner, OptSpaces
         if Scanner, IsComment($) then
            Scanner, Comment
            Scanner, OptMiscs
         elseif Scanner, IsPI($) then
            Scanner, PI
            Scanner, OptMiscs
         else skip end
      end
      
% %% Document Type Definition

% [28] doctypedecl ::=  '<!DOCTYPE' S Name (S ExternalID)? S? ('[' (markupdecl | PEReference | S)*
%                       ']' S?)? '>'
%                                       [ VC: Root Element Type ]
% [29] markupdecl  ::=  elementdecl | AttlistDecl | EntityDecl | NotationDecl | PI | Comment 
%                                       [ VC: Proper Declaration/PE Nesting ]
%                                       [ WFC: PEs in Internal Subset ]

      meth IsDoctypedecl($)
         Scanner, IsPrefix("<!DOCTYPE" $)
      end
      meth Doctypedecl
         N ExtID in
	 Scanner, Str("<!DOCTYPE")
	 Scanner, Spaces
	 Scanner, NameA(?N)
         Scanner, OptSpaces
         if Scanner, IsExternalID($) then
            Scanner, ExternalID(?ExtID)
            Scanner, OptSpaces
         else
	    ExtID=unit
         end
         Scanner, putToken('<!DOCTYPE' N#ExtID)
         if Scanner, topC($) == &[ then
            Scanner, Chr(&[)
            Scanner, substPEOn
            Scanner, OptSpaces
            Scanner, DtdItems
            Scanner, substPEOff
	    Scanner, Chr(&])
            Scanner, OptSpaces
	 end
	 Scanner, Chr(&>)
         Scanner, putToken1('>')
      end
      meth DtdItems
         if Scanner, topC($) == &] then skip
         else
            Continue =
            if Scanner, IsPEReference($) then
               Scanner, IncludePEReference true
            else
               Scanner, Markupdecl($)
            end
         in
            {Skip} %% compiler bug
            if Continue==false then skip
            else
               Scanner, OptSpaces
               Scanner, DtdItems
            end
         end
      end
      meth IncludePEReference
         X Y U Z in
         X = Scanner, PEReferenceA($)
         Y = Refs, lookupPE(X $)
         %%Scanner, putToken('include' X#Y)
         %%{System.printInfo "@url: "} {System.showInfo @url}
         %%{System.printInfo "PE: "} {System.show Y}
         case Y of 'VAL'(VS) then
            U=@url Z={VirtualString.toString VS}
         elseif @includeExternalPEs==false then
            %% hack! base url needs to be passed in when parsing documents
            U='' Z=nil
         elsecase Y of 'ID'(system:SystemID) then
            U={MkUrl @url SystemID}
            Z={LoadUrlWithTail U nil}
         [] 'ID'(public:_ system:SystemID) then
            U={MkUrl @url SystemID}
            Z={LoadUrlWithTail U nil}
         end
         Scanner, PushBuffer
         Scanner, InitBuffer(Z url:U)
         Scanner, extSubset
         Scanner, PopBuffer
      end
      %% markupdecl ends with "]" or end-of-file
      meth Markupdecl($)
         if Scanner, IsComment($) then
            Scanner, Comment true
         elseif Scanner, IsPI($) then
            %% what about PERefs in PIs?
            Scanner, PI true
         elseif Scanner, IsNotationDecl($) then
            %% what about PERefs in NotationDecls?
	    Scanner, NotationDecl true
         elseif Scanner, IsEntityDecl($) then
            %% what about PERefs in EntityDecls?
            Scanner, EntityDecl true
         elseif Scanner, IsAttlistDecl($) then
            %% what about PERefs in AttlistDecls?
	    Scanner, AttlistDecl true
         elseif Scanner, IsElementDecl($) then
            %% what about PERefs in ElementDecls?
	    Scanner, ElementDecl true
	 else false end
      end
      
% %% External Subset

% [30] extSubset     ::=  TextDecl? extSubsetDecl
% [31] extSubsetDecl ::=  ( markupdecl | conditionalSect | PEReference | S )*

      %% extSubset ends with end-of-file 
      meth extSubset
         Scanner, substPEOn
         if Scanner, IsTextDecl($) then
            Scanner, TextDecl
         end
         Scanner, ExtSubsetDecl
      end
      %% extSubsetDecl ends with "]]>" or end-of-file
      meth ExtSubsetDecl
         Scanner, OptSpaces
         if Scanner, IsPEReference($) then
            Scanner, IncludePEReference
            Scanner, ExtSubsetDecl
         elseif Scanner, IsConditionalSect($) then
            Scanner, ConditionalSect
	    Scanner, ExtSubsetDecl
         elseif Scanner, Markupdecl($) == false then skip
         else
	    Scanner, ExtSubsetDecl
         end
      end
      
% %%
% %% 2.9 Standalone Document Declaration
% %%

% [32] SDDecl ::=  S 'standalone' Eq (("'" ('yes' | 'no') "'") | ('"' ('yes' | 'no') '"')) 
%                                       [ VC: Standalone Document Declaration ]
      
      meth IsSDDecl($)
         Scanner, IsPrefix("standalone" $)
      end
      meth SDDecl($)
         Scanner, Str("standalone")
         Scanner, Eq
         Scanner, YesNoLiteral($)
      end
      meth YesNoLiteral($)
         case Scanner, Literal($)
         of "yes" then yes
         [] "no"  then no
         else
            Scanner, reportErr('"yes" or "no" expected') dummy
         end
      end

% %%
% %% 2.10 White Space Handling
% %%

      %% not supported
      
% %%
% %% 2.11 End-of-Line Handling
% %%

      %% not supported
      
% %%
% %% 2.12 Language Identification
% %%

% [33] LanguageID ::=  Langcode ('-' Subcode)*
% [34]   Langcode ::=  ISO639Code |  IanaCode |  UserCode
% [35] ISO639Code ::=  ([a-z] | [A-Z]) ([a-z] | [A-Z])
% [36]   IanaCode ::=  ('i' | 'I') '-' ([a-z] | [A-Z])+
% [37]   UserCode ::=  ('x' | 'X') '-' ([a-z] | [A-Z])+
% [38]   Subcode  ::=  ([a-z] | [A-Z])+
      
      meth LanguageID($)
         Lang = Scanner, Langcode($) in
         Lang # Scanner, Subcodes($)
      end
      meth Langcode($)
         if Scanner, IsIanaCode($) then
            &i|&-|Scanner, IanaCode($)
         elseif Scanner, IsUserCode($) then
            &x|&-|Scanner, UserCode($)
         else
            Scanner, ISO639Code($)
         end
      end
      meth ISO639Code(?Code)
         C1 = Scanner, getC($)
         C2 = Scanner, getC($)
      in
         if {IsAZaz C1} andthen
            {IsAZaz C2} then
            Code = [C1 C2]
         else
            Scanner, reportErr('language code expected')
         end
      end
      meth IsIanaCode($)
         Scanner, IsPrefix("i-" $) orelse
         Scanner, IsPrefix("I-" $) 
      end
      meth IanaCode(?Code)
         Scanner, skipC   %  &i | &I
         Scanner, Chr(&-)
         Code = Scanner, `[A-Za-z]+`($)
      end
      meth IsUserCode($)
         Scanner, IsPrefix("x-" $) orelse
         Scanner, IsPrefix("X-" $)
      end
      meth UserCode(?Code)
         Scanner, skipC   %  &x | &X
         Scanner, Chr(&-)
         Code = Scanner, `[A-Za-z]+`($)
      end
      meth IsSubcodes($)
         Scanner, topC($) == &-
      end
      meth Subcodes($)
         if Scanner, IsSubcodes($)==false then nil
         else
            Sub = Scanner, Subcode($) in
            Sub # Scanner, Subcodes($)
         end
      end
      meth Subcode($)
         Scanner, `[A-Za-z]+`($)
      end
      meth `[A-Za-z]+`($)
         C = Scanner, topC($) in
         if {IsAZaz C} then
            Scanner, skipC
            C | Scanner, `[A-Za-z]*`($)
         else
            Scanner, reportErr('basic letter expected [A-Za-z]') dummy
         end
      end
      meth `[A-Za-z]*`($)
         C = Scanner, topC($) in
         if {IsAZaz C}==false then nil
         else
            Scanner, skipC
            C | Scanner, `[A-Za-z]*`($)
         end
      end
      
% %%%
% %%% 3. Logical Structures
% %%%

% %% Element

% [39] element ::=  EmptyElemTag
%                   | STag content ETag
%                                       [ WFC: Element Type Match ]
%                                       [ VC: Element Valid ]

      meth IsElement($)
         Scanner, IsSTag($)
      end
      meth Element
         if Scanner, IsSTag($) == false then
            Scanner, reportErr('start tag expected')
         else
            Empty in
            Scanner, STag(?Empty)
            if Empty==false then
               Scanner, Content
               Scanner, ETag
            end
         end
      end
      
% %%
% %% 3.1 Start-Tags, End-Tags, and Empty-Element Tags
% %%

% %% Start-tag

% [40]      STag ::=  '<' Name (S Attribute)* S? '>'
%                                       [ WFC: Unique Att Spec ]
% [41] Attribute ::=  Name Eq AttValue
%                                       [ VC: Attribute Value Type ]
%                                       [ WFC: No External Entity References ]
%                                       [ WFC: No < in Attribute Values ]
      
      meth IsSTag($)
         Scanner, topC($) == &< andthen
         Scanner, IsPrefix("</" $) == false
      end
      meth STag(?Empty) % including EmptyElemTag
         N As in
         Scanner, Chr(&<)
         Scanner, NameA(?N)
         Scanner, OptSpaces
         Scanner, Attributes(?As) 
         Empty = (Scanner, topC($) == &/)
         if Empty then
            Scanner, Chr(&/)
         end
         Scanner, Chr(&>)
         Scanner, putToken(startElement N#As)
         if Empty then
            Scanner, putToken(endElement N)
         end
      end
      meth Attributes($)
         C = Scanner, topC($) in
         if C==&/ orelse C==&> then nil
         else
            Pair = Scanner, Attribute($) in
            Pair | Scanner, Attributes($)
         end
      end
      meth Attribute(?Pair)
         N#V = Pair in
         Scanner, NameA(?N)
         Scanner, Eq
         Scanner, AttValue(?V)
         Scanner, OptSpaces
      end
      
% %% End-tag
      
% [42] ETag ::=  '</' Name S? '>'
      
      meth IsETag($)
         Scanner, IsPrefix("</" $)
      end
      meth ETag
         N in
         Scanner, Str("</")
         Scanner, NameA(?N)
         Scanner, OptSpaces
         Scanner, Chr(&>)
         Scanner, putToken(endElement N)
      end
      
% %% Content of Elements

% [43] content ::=  (element | CharData | Reference | CDSect | PI | Comment)*

      %% content ends with "</...>" or end-of-file
      meth Content
         if Scanner, atEnd($) orelse
            Scanner, IsPrefix("</" $) then skip
         else
            Scanner, ContentItem
            Scanner, Content
         end
      end
      meth ContentItem
         if Scanner, IsComment($) then
            Scanner, Comment
         elseif Scanner, IsPI($) then
            Scanner, PI
         elseif Scanner, IsCDSect($) then
            Scanner, CDSect
         elseif Scanner, IsReference($) then
            Scanner, Reference
         elseif Scanner, IsElement($) then
            Scanner, Element
         else
            Scanner, CharData
         end
      end

% %% Tags for Empty Elements

% [44] EmptyElemTag ::=  '<' Name (S Attribute)* S? '/>'
%                                       [ WFC: Unique Att Spec ]

      %% integrated in STag

% %%
% %% 3.2 Element Type Declarations
% %%
                
% [45] elementdecl ::=  '<!ELEMENT' S Name S contentspec S? '>'
%                                       [ VC: Unique Element Type Declaration ]
% [46] contentspec ::=  'EMPTY' | 'ANY' | Mixed | children 

      meth IsElementDecl($)
         Scanner, IsPrefix("<!ELEMENT" $)
      end
      meth ElementDecl
         N Model in
         Scanner, Str("<!ELEMENT")
         Scanner, Spaces
         Scanner, NameA(?N)
         Scanner, Spaces
         Scanner, ContentSpec(?Model)
         Scanner, OptSpaces
         Scanner, Chr(&>)
         Scanner, putToken('ELEMENT' 'ELEMENT'(N Model))
      end
      meth ContentSpec($)
         if Scanner, IsPrefix("EMPTY" $) then
            Scanner, Str("EMPTY") 'EMPTY'
         elseif Scanner, IsPrefix("ANY" $) then
            Scanner, Str("ANY") 'ANY'
         elseif Scanner, IsMixed($) then
            case Scanner, Mixed($) of nil then 'PCDATA'
            [] ElemNames then mixed(ElemNames) end
         else
            children(Scanner, Children($))
         end
      end

% %%
% %% 3.2.1 Element Content
% %%
      
% %% Element-content Models

% [47] children ::=  (choice | seq) ('?' | '*' | '+')?
% [48]       cp ::=  (Name | choice | seq) ('?' | '*' | '+')?
% [49]   choice ::=  '(' S? cp ( S? '|' S? cp )* S? ')'
%                                       [ VC: Proper Group/PE Nesting ]
% [50] seq ::=  '(' S? cp ( S? ',' S? cp )* S? ')'
%                                       [ VC: Proper Group/PE Nesting ]

      meth OperatorChr($)
         case Scanner, topC($)
         of &? then Scanner, Chr(&?) '?'
         [] &* then Scanner, Chr(&*) '*'
         [] &+ then Scanner, Chr(&+) '+'
         else unit
         end 
      end
      meth Children($)
         G  = Scanner, Group($) in
         case Scanner, OperatorChr($)
         of unit then G
         [] Op   then Op(G)
         end
      end
      meth Cp($)
         if Scanner, IsName($) then
            N = Scanner, NameA($) in
            case Scanner, OperatorChr($)
            of unit then N
            [] Op   then Op(N)
            end
         else
            G = Scanner, Group($) in
            case Scanner, OperatorChr($)
            of unit then G
            [] Op   then Op(G)
            end
         end
      end
      /*
      meth IsGroupOrPERef($)
         Scanner, IsGroup($) orelse
         Scanner, IsPEReference($)
      end
      meth IsGroup($)
         Scanner, topC($) == &(
      end
      */
      meth Group($) %% or PERef
         if Scanner, IsPEReference($) then
            N = Scanner, PEReferenceA($) in
            '%REF'(N)
         else
            Cp1 in
            Scanner, Chr(&()
            Scanner, OptSpaces
            Scanner, Cp(?Cp1)
            Scanner, OptSpaces
            case Scanner, topC($) of &| then
               'choice'(Cp1 | Scanner, ChoiceCps($))
            [] &, then
               'seq'(Cp1 | Scanner, SeqCps($))
            else
               Scanner, Chr(&))
               Cp1
            end
         end
      end
      meth ChoiceCps($)
         case Scanner, topC($) of &) then
            Scanner, Chr(&)) nil
         else
            Cp1 in
            Scanner, Chr(&|)
            Scanner, OptSpaces
            Scanner, Cp(?Cp1)
            Scanner, OptSpaces
            Cp1 | Scanner, ChoiceCps($)
         end
      end
      meth SeqCps($)
         case Scanner, topC($) of &) then
            Scanner, Chr(&)) nil
         else
            Cp1 in
            Scanner, Chr(&,)
            Scanner, OptSpaces
            Scanner, Cp(?Cp1)
            Scanner, OptSpaces
            Cp1 | Scanner, SeqCps($)
         end
      end

% %%   
% %% 3.2.2 Mixed Content
% %%

% %% Mixed-content Declaration

% [51] Mixed ::=  '(' S? '#PCDATA' (S? '|' S? Name)* S? ')*' 
%                 | '(' S? '#PCDATA' S? ')' 
%                                       [ VC: Proper Group/PE Nesting ]
%                                       [ VC: No Duplicate Types ]
      
      meth IsMixed($)
         Scanner, topC($) == &( andthen
         local _|S=@s in
            {List.isPrefix "#PCDATA" {List.dropWhile S IsSpaceChar}}
         end
      end
      meth Mixed(?ElementNames)
         Scanner, Chr(&()
         Scanner, OptSpaces
         Scanner, Str("#PCDATA")
         Scanner, OptSpaces
         Scanner, MixedNames(?ElementNames)
         if Scanner, IsPrefix(")*" $) then
            Scanner, Str(")*")
         else
            Scanner, Chr(&))
         end
      end
      meth MixedNames(?Ns)
         if Scanner, topC($) \= &| then Ns=nil
         else
            N|Nr = Ns in
            Scanner, Chr(&|)
            Scanner, OptSpaces
            Scanner, NameA(?N)
            Scanner, OptSpaces
            Scanner, MixedNames(?Nr)
         end
      end
      
% %%
% %% 3.3 Attribute-List Declarations
% %%

% %% Attribute-list Declaration

% [52] AttlistDecl ::=  '<!ATTLIST' S Name AttDef* S? '>'
% [53]      AttDef ::=  S Name S AttType S DefaultDecl

      meth IsAttlistDecl($)
         Scanner, IsPrefix("<!ATTLIST" $)
      end
      meth AttlistDecl
         N Defs in
         Scanner, Str("<!ATTLIST")
         Scanner, Spaces
	 Scanner, NameA(?N)
	 Scanner, OptSpaces
         Scanner, AttDefs(?Defs)
	 Scanner, Chr(&>)
         Scanner, putToken('ATTLIST' 'ATTLIST'(N Defs))
      end
      meth AttDefs($)
	 if Scanner, topC($) == &> then nil
         else
	    Def = Scanner, AttDef($) in
	    Def | Scanner, AttDefs($)
	 end
      end
      meth AttDef($)
         N Type Default in
         Scanner, NameA(?N)
         Scanner, Spaces
         Scanner, AttType(?Type)
         Scanner, Spaces
         Scanner, DefaultDecl(?Default)
         Scanner, OptSpaces
         N#'#'(Type Default)
      end
      
% %%
% %% 3.3.1 Attribute Types
% %%

% %% Attribute Types

% [54]       AttType ::=  StringType | TokenizedType | EnumeratedType 
% [55]    StringType ::=  'CDATA'
% [56] TokenizedType ::=  'ID'
%                                       [ VC: ID ]
%                                       [ VC: One ID per Element Type ]
%                                       [ VC: ID Attribute Default ]
%                         | 'IDREF'
%                                       [ VC: IDREF ]
%                         | 'IDREFS'
%                                       [ VC: IDREF ]
%                         | 'ENTITY'
%                                       [ VC: Entity Name ]
%                         | 'ENTITIES'
%                                       [ VC: Entity Name ]
%                         | 'NMTOKEN'
%                                       [ VC: Name Token ]
%                         | 'NMTOKENS'
%                                       [ VC: Name Token ]

      meth AttType($) 
         if Scanner, IsPrefix("CDATA" $) then
            Scanner, Str("CDATA") 'CDATA'
         elseif Scanner, IsPrefix("IDREFS" $) then
            Scanner, Str("IDREFS")   'IDREFS'
         elseif Scanner, IsPrefix("IDREF" $) then
            Scanner, Str("IDREF")    'IDREF'
         elseif Scanner, IsPrefix("ID" $) then
            Scanner, Str("ID")       'ID'
         elseif Scanner, IsPrefix("ENTITIES" $) then
            Scanner, Str("ENTITIES") 'ENTITIES'
         elseif Scanner, IsPrefix("ENTITY" $) then
            Scanner, Str("ENTITY")   'ENTITY'
         elseif Scanner, IsPrefix("NMTOKENS" $) then
            Scanner, Str("NMTOKENS") 'NMTOKENS'
         elseif Scanner, IsPrefix("NMTOKEN" $) then
            Scanner, Str("NMTOKEN")  'NMTOKEN'
         else
            Scanner, EnumeratedType($)
         end
      end
      
% %% Enumerated Attribute Types

% [57] EnumeratedType ::=  NotationType | Enumeration 
% [58]   NotationType ::=  'NOTATION' S '(' S? Name (S? '|' S? Name)* S? ')' 
%                                       [ VC: Notation Attributes ]
% [59]    Enumeration ::=  '(' S? Nmtoken (S? '|' S? Nmtoken)* S? ')'
%                                       [ VC: Enumeration ]
      
      meth EnumeratedType($)
         if Scanner, IsNotationType($) then
            Ns = Scanner, NotationType($) in
            'NOTATION'(Ns)
         else
            Ns = Scanner, Enumeration($) in
            'ENUMERATION'(Ns)
         end
      end
      meth IsNotationType($)
         Scanner, IsPrefix("NOTATION" $)
      end
      meth NotationType(?Ns)
         N|Nr = Ns in
         Scanner, Str("NOTATION")
         Scanner, Spaces
         Scanner, Chr(&()
         Scanner, OptSpaces
         Scanner, NameA(?N)
         Scanner, OptSpaces
         Scanner, ChoiceNames(?Nr)
         Scanner, Chr(&))
      end
      meth ChoiceNames($)
         if Scanner, topC($) \= &| then nil
         else
            Scanner, Chr(&|)
            Scanner, OptSpaces
            N in
            Scanner, NameA(?N)
            Scanner, OptSpaces
            N | Scanner, ChoiceNames($)
         end
      end
      meth Enumeration(?Ns)
         N|Nr = Ns in
         Scanner, Chr(&()
         Scanner, OptSpaces
         Scanner, Nmtoken(?N)
         Scanner, OptSpaces
         Scanner, ChoiceNmtokens(?Nr)
         Scanner, Chr(&))
      end
      meth ChoiceNmtokens($)
         if Scanner, topC($) \= &| then nil
         else
            Scanner, Chr(&|)
            Scanner, OptSpaces
            N in
            Scanner, Nmtoken(?N)
            Scanner, OptSpaces
            N | Scanner, ChoiceNmtokens($)
         end
      end
 
% %%
% %% 3.3.2 Attribute Defaults
% %%

% %% Attribute Defaults

% [60] DefaultDecl ::=  '#REQUIRED' | '#IMPLIED' 
%                       | (('#FIXED' S)? AttValue)
%                                       [ VC: Required Attribute ]
%                                       [ VC: Attribute Default Legal ]
%                                       [ WFC: No < in Attribute Values ]
%                                       [ VC: Fixed Attribute Default ]
      
      meth DefaultDecl($)
         if Scanner, IsPrefix("#REQUIRED" $) then
	    Scanner, Str("#REQUIRED") '#REQUIRED'
         elseif Scanner, IsPrefix("#IMPLIED" $) then
            Scanner, Str("#IMPLIED")  '#IMPLIED'
         elseif Scanner, IsPrefix("#FIXED" $) then
	    Val in
	    Scanner, Str("#FIXED")
            Scanner, Spaces
            Scanner, AttValue(?Val)
	    '#FIXED'(Val)
         else
            Val in
            Scanner, AttValue(?Val)
            unit(Val)
         end
      end
      
% %%
% %% 3.3.3 Attribute-Value Normalization
% %%

% %%
% %% 3.4 Conditional Sections
% %%

% %% Conditional Section

% [61]    conditionalSect ::=  includeSect | ignoreSect 
% [62]        includeSect ::=  '<![' S? 'INCLUDE' S? '[' extSubsetDecl ']]>' 
% [63]         ignoreSect ::=  '<![' S? 'IGNORE' S? '[' ignoreSectContents* ']]>'
% [64] ignoreSectContents ::=  Ignore ('<![' ignoreSectContents ']]>' Ignore)*
% [65]             Ignore ::=  Char* - (Char* ('<![' | ']]>') Char*)
      
      meth IsConditionalSect($)
         Scanner, IsPrefix("<![" $)
      end
      meth ConditionalSect
	 Scanner, Str("<![")
         Scanner, OptSpaces
	 if Scanner, IsPrefix("INCLUDE" $) then
	    Scanner, IncludeSect
         elseif Scanner, IsPrefix("IGNORE" $) then
	    Scanner, IgnoreSect
	 else
	    Scanner, reportErr('"INCLUDE" or "IGNORE" expected')
	 end
      end
      meth IncludeSect
	 Scanner, Str("INCLUDE")
	 Scanner, OptSpaces
	 Scanner, Chr(&[)
	 Scanner, ExtSubsetDecl
	 Scanner, Str("]]>")
      end
      meth IgnoreSect
         Scanner, substPEOff
	 Scanner, Str("IGNORE")
	 Scanner, OptSpaces
	 Scanner, Chr(&[)
         Scanner, IgnoreSectContents
         Scanner, Str("]]>")
         Scanner, substPEOn
      end
      meth IgnoreSectContents
	 Scanner, Ignore
	 if Scanner, IsPrefix("<![" $)==false then skip
	 else
	    Scanner, Str("<![")
	    Scanner, IgnoreSectContents
	    Scanner, Str("]]>")
	    Scanner, IgnoreSectContents
	 end
      end
      meth Ignore
	 if Scanner, IsPrefix("<![" $) then skip
	 elseif Scanner, IsPrefix("]]>" $) then skip
	 else
	    Scanner, skipC
	    Scanner, Ignore
	 end
      end
      
% %%% 
% %%% 4. Physical Structures
% %%%

% %%
% %% 4.1 Character and Entity References
% %%

% %% Character Reference

% [66] CharRef ::=  '&#' [0-9]+ ';' 
%                   | '&#x' [0-9a-fA-F]+ ';'
%                                       [ WFC: Legal Character ]
      
      meth CharRef($)
         C in
         Scanner, Str("&#")
         case Scanner, ReadUntilC(&; $)
         of &x|Hex then Scanner, Hex2Int(Hex ?C)
         [] Dec    then Scanner, Dec2Int(Dec ?C)
         end
         if {IsInt C} andthen {IsChar C}==false then
            %%Scanner, CharRefErr([hint#'value out of bounds' value#{VS2A C}])
            error({VirtualString.toString "&#"#C#";"})
         else C end
      end
      meth Dec2Int(Dec ?C)
         Dec1 = {DropZeroes Dec} in
         if {String.isInt Dec1} then
            C={StringToInt Dec1}
         else
            %%Scanner, CharRefErr([decimal#{VS2A Dec}])
            C=error({Append &&|&#|Dec ";"})
         end
      end
      meth Hex2Int(Hex ?C)
         try
            C={Compiler.virtualStringToValue &&|&\\|&x|Hex}
         catch _ then
            %%Scanner, CharRefErr([hex#{VS2A Hex}])
            C=error({Append &&|&#|&x|Hex ";"})
         end
      end
      meth CharRefErr(Hints)
         Scanner, reportErr('illegal character reference' hints:Hints)
      end
      
% %% Entity Reference

% [67]   Reference ::=  EntityRef | CharRef
% [68]   EntityRef ::=  '&' Name ';'
%                                       [ WFC: Entity Declared ]
%                                       [ VC: Entity Declared ]
%                                       [ WFC: Parsed Entity ]
%                                       [ WFC: No Recursion ]
% [69] PEReference ::=  '%' Name ';'
%                                       [ VC: Entity Declared ]
%                                       [ WFC: No Recursion ]
%                                       [ WFC: In DTD ]

      meth IsReference($)
         Scanner, topC($) == &&
      end
      
      meth Reference
         if Scanner, IsPrefix("&#" $) then
            case Scanner, CharRef($)
            of error(S) then
               Scanner, putToken(characters S)
            [] C then
               Scanner, putToken(characters {Char.toAtom C})
            end
         else
            N in
            Scanner, EntityRef(?N)
            Scanner, putToken(entityRef N)
         end
      end
      meth EntityRef(?NameS)
         Scanner, Chr(&&)
         Scanner, RefName(?NameS)
      end
      meth IsPEReference($)
         case @s of &%|C|_ then
            {IsSpaceChar C}==false
         else false end
      end
      meth PEReference(?NameS)
         Scanner, Chr(&%)
         Scanner, RefName(?NameS)
      end
      meth PEReferenceA($)
         N = Scanner, PEReference($) in
         {StringToAtom N}
      end
      meth RefName(?NameS)
	 Scanner, ReadUntilC(&; ?NameS)
      end
      
% %%
% %% 4.2 Entity Declarations
% %%

% %% Entity Declaration

% [70] EntityDecl ::=  GEDecl | PEDecl
% [71]     GEDecl ::=  '<!ENTITY' S Name S EntityDef S? '>'
% [72]     PEDecl ::=  '<!ENTITY' S '%' S Name S PEDef S? '>'
% [73]  EntityDef ::=  EntityValue | (ExternalID NDataDecl?)
% [74]      PEDef ::=  EntityValue | ExternalID

      meth IsEntityDecl($)
         Scanner, IsPrefix("<!ENTITY" $)
      end
      meth EntityDecl
         Scanner, substPEOff
         Scanner, Str("<!ENTITY")
         Scanner, Spaces
         if Scanner, IsPEDecl($) then 
            Scanner, PEDecl
         else
            Scanner, GEDecl
         end
      end
      meth GEDecl
         N Def in
         Scanner, NameA(?N)
         Scanner, substPEOn
         Scanner, Spaces
         Scanner, EntityDef(?Def)
         Scanner, OptSpaces
         Scanner, Chr(&>)
         Refs, enterGE(N Def)
         Scanner, putToken('ENTITY' '\&'(N Def))
      end
      meth IsPEDecl($)
         Scanner, topC($) == &%
      end
      meth PEDecl
         N Def in
         Scanner, Chr(&%)
         Scanner, Spaces
         Scanner, NameA(?N)
         Scanner, substPEOn
         Scanner, Spaces
         Scanner, PEDef(?Def)
         Scanner, OptSpaces
         Scanner, substPEOff
         Scanner, Chr(&>)
         Refs, enterPE(N Def)
         Scanner, substPEOn 
         Scanner, putToken('ENTITY' '%'(N Def))
      end
      meth EntityDef($)
         if Scanner, IsEntityValue($) then
            Val = Scanner, EntityValue($) in
            'VAL'(Val)
         else
            ID = Scanner, ExternalID($) in
            Scanner, OptSpaces
            if Scanner, IsNDATA($) then
               N in
               Scanner, NDATA(?N)
               Scanner, OptSpaces
               {AdjoinAt ID ndata N}
            else
               ID
            end
         end
      end
      meth PEDef($)
         if Scanner, IsEntityValue($) then
            Val = Scanner, EntityValue($) in
            'VAL'(Val)
         else
            Scanner, ExternalID($)
         end
      end
      
% %%
% %% 4.2.1 Internal Entities
% %%

% %%
% %% 4.2.2 External Entities
% %%

% %% External Entity Declaration

% [75] ExternalID ::=  'SYSTEM' S SystemLiteral
%                      | 'PUBLIC' S PubidLiteral S SystemLiteral 
% [76] NDataDecl ::=  S 'NDATA' S Name
%                                       [ VC: Notation Declared ]

      meth IsPUBLICorSYSTEM($)
         Scanner, IsSYSTEM($) orelse
         Scanner, IsPUBLIC($)
      end
      meth IsExternalID($)
         Scanner, IsPUBLICorSYSTEM($)
      end
      %% ExternalID ends with "[", ">" or "NDATA"
      meth ExternalID($)
         if Scanner, IsSYSTEM($) then
            'ID'(system: Scanner, SYSTEM($))
         elseif Scanner, IsPUBLIC($) then
            Pub Sys in
            Scanner, PUBLIC(?Pub ?Sys)
            'ID'(public:Pub system:Sys)
         else
            Scanner, reportErr('external ID: "SYSTEM" or "PUBLIC" expected')
            dummy
         end
      end
      meth IsSYSTEM($)
         Scanner, IsPrefix("SYSTEM" $)
      end
      meth SYSTEM(?Sys)
         Scanner, Str("SYSTEM")
         Scanner, Spaces
         Scanner, SystemLiteral(?Sys)
         Scanner, OptSpaces
      end
      meth IsPUBLIC($)
         Scanner, IsPrefix("PUBLIC" $)
      end
      %% PublicID ends with ">"
      meth PUBLIC(?Pub ?Sys)
         Scanner, Str("PUBLIC")
         Scanner, Spaces
         Scanner, PubidLiteral(?Pub)
         Scanner, OptSpaces
         if Scanner, topC($) \= &> then
            Scanner, SystemLiteral(?Sys)
            Scanner, OptSpaces
         else
            Sys = unit
         end
      end
      meth IsNDATA($)
         Scanner, IsPrefix("NDATA" $)
      end
      meth NDATA(?N)
         Scanner, Str("NDATA")
         Scanner, Spaces
         Scanner, Name(?N)
      end
      
% %%
% %% 4.3 Parsed Entities
% %%

% %%
% %% 4.3.1 The Text Declaration
% %%

% [77] TextDecl ::=  '<?xml' VersionInfo? EncodingDecl S? '?>'

      meth IsTextDecl($)
         Scanner, IsPrefix("<?xml" $)
      end
      meth TextDecl
         Ver Enc in
         Scanner, Str("<?xml")
         Scanner, Spaces
         if Scanner, IsVersionInfo($) then
            Scanner, VersionInfo(?Ver)
         else
            Ver = unit
         end
         if Scanner, IsEncodingDecl($) then
            Scanner, EncodingDecl(?Enc)
         else
            Scanner, reportErr('encoding declaration expected')
         end
         Scanner, OptSpaces
         Scanner, Str("?>")
      end
      
% %%
% %% 4.3.2 Well-Formed Parsed Entities
% %%

% %% Well-Formed External Parsed Entity

% [78] extParsedEnt ::=  TextDecl? content
% [79]        extPE ::=  TextDecl? extSubsetDecl

      %% extParsedEnt ends with end-of-file
      meth extParsedEnt
         if Scanner, IsTextDecl($) then
            Scanner, TextDecl
         end
         Scanner, Content
      end
      %% extPE ends with "]]>" or end-of-file
      meth ExtPE
         if Scanner, IsTextDecl($) then
            Scanner, TextDecl
         end
         Scanner, ExtSubsetDecl
      end
         
% %%
% %% 4.3.3 Character Encoding in Entities
% %%

% %% Encoding Declaration

% [80] EncodingDecl ::=  S 'encoding' Eq ('"' EncName '"' |  "'" EncName "'" ) 
% [81]      EncName ::=  [A-Za-z] ([A-Za-z0-9._] | '-')*
% /* Encoding name contains only Latin characters */

      meth IsEncodingDecl($)
         Scanner, IsPrefix("encoding" $)
      end
      meth EncodingDecl(?EncNameS)
         Scanner, Str("encoding")
         Scanner, Eq
         Scanner, Literal(?EncNameS)
         Scanner, CheckEncName(EncNameS)
      end
      meth CheckEncName(EncNameS)
         case EncNameS of nil then
            Scanner, reportErr('encoding name empty')
         [] C|R then
            if {IsAZaz C}==false then
               Scanner, reportErr('encoding name must begin with [A-Za-z]')
            elseif {All R IsEncNameChar}==false then
               Scanner, reportErr('encoding name contains illegal character'
                                  hints:[allowed#'[A-Za-z0-9._] | "-"'
                                         found  #{StringToAtom R}])
            end
         end
      end
      
% %%
% %% 4.4 XML Processor Treatment of Entities and References
% %% 4.5 Construction of Internal Entity Replacement Text
% %%

% %%
% %% 4.6 Predefined Entities
% %%

% <!ENTITY lt     "&#38;#60;"> 
% <!ENTITY gt     "&#62;"> 
% <!ENTITY amp    "&#38;#38;"> 
% <!ENTITY apos   "&#39;"> 
% <!ENTITY quot   "&#34;">
         
% %%
% %% 4.7 Notation Declarations
% %%

% %% Notation Declarations

% [82] NotationDecl ::=  '<!NOTATION' S Name S (ExternalID |  PublicID) S? '>'
% [83]     PublicID ::=  'PUBLIC' S PubidLiteral 
      
      meth IsNotationDecl($)
         Scanner, IsPrefix("<!NOTATION" $)
      end
      meth NotationDecl
         N ID in
         Scanner, Str("<!NOTATION")
         Scanner, Spaces
         Scanner, NameA(?N)
         Scanner, Spaces
         Scanner, ExternalOrPublicID(?ID)
         Scanner, OptSpaces
         Scanner, Chr(&>)
         Scanner, putToken('NOTATION' 'NOTATION'(N ID))
      end
      meth ExternalOrPublicID($)
         if Scanner, IsSYSTEM($) then
            Scanner, ExternalID($)
         else
            Pub Sys in
            Scanner, PUBLIC(?Pub ?Sys)
            if Sys==unit
            then 'ID'(public:Pub)
            else 'ID'(public:Pub system:Sys)           
            end
         end
      end

% %%
% %% 4.8 Document Entity
% %%

% %%%
% %%% 5. Conformance
% %%%
% %%
% %% 5.1 Validating and Non-Validating Processors
% %% 5.2 Using XML Processors
% %%

   end % of class Scanner

   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% scanner errors %%%%%%%%%%%%%%%%%%%%%%%%%
   {Error.registerFormatter scanner
    fun {$ E}
       T = 'scanner error' 
       BugReport = 'please send a bug report to afranke@ags.uni-sb.de' 
    in 
       case E of scanner(internal X) then 
          error(kind:T msg:'Internal scanner error' 
                items:[hint(l:'Additional information' m:oz(X))
                       line(BugReport)])
       [] scanner(externalPE(N Def) pos:Pos...) then
          error(kind:T msg:'unsupported external parameter reference'
                items: [Pos unit
                        hint(l:name   m:oz(N))
                        unit
                        hint(l:'PUBLIC' m:{CondSelect Def public "<unknown>"})
                        hint(l:'SYSTEM' m:{CondSelect Def system "<unknown>"})
                       ])
       [] scanner(Msg hints:Hints pos:Pos...) then
          error(kind:T msg:Msg
                items: Pos|unit|{Map Hints fun {$ X#Y} hint(l:X m:oz(Y)) end})
       else
          error(kind:T items:[line(oz(E))])
       end 
    end}
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   %%
   %% exports
   %%
   local
      /*
      local
         fun  {GetDebug} {Boot_Debug.getRaiseOnBlock {Thread.this}} end
         proc {DebugOn}  {Boot_Debug.setRaiseOnBlock {Thread.this} true}  end
         proc {DebugOff} {Boot_Debug.setRaiseOnBlock {Thread.this} false} end
      in
         proc {Nonblocking_ P}
            if {GetDebug} then {P}
            else {DebugOn} {P} {DebugOff} end
         end
      end
      */
      proc {Nonblocking P} {P} end % switched off (critical: including DTDs)
      %%
      VS2S = VirtualString.toString
      proc {LoadUrl Url ?S}
         F={New Open.file init(url:Url)} in
         {F read(list:?S size:all)}
         {F close}
      end
      proc {LoadFile FileName ?S}
         F={New Open.file init(name:FileName)} in
         {F read(list:?S size:all)}
         {F close}
      end 
   in
      %% DTDs
      local
         proc {ScanDTD S BaseUrl ?Tokens}
            {Nonblocking
             proc {$} 
                MyScanner = {New Scanner init(S url:BaseUrl)} in
                {MyScanner extSubset}
                {MyScanner getTokens(?Tokens)}
             end}
         end
      in
         DTD = dtd(scan:       ScanDTD
                   scanString: fun {$ S}  {ScanDTD S           ''} end
                   scanVS:     fun {$ VS} {ScanDTD {VS2S VS}   ''} end
                   scanFile:   fun {$ F}  {ScanDTD {LoadFile F} F} end
                   scanURL:    fun {$ U}  {ScanDTD {LoadUrl  U} U} end
                  )
      end
      
      %% XML documents
      local
         proc {ScanXML S BaseUrl ?Tokens}
            {Nonblocking
             proc {$}
                MyScanner = {New Scanner init(S url:BaseUrl)} in
                {MyScanner document}
                {MyScanner getTokens(?Tokens)}
             end}
         end
      in
         Document = doc(scan:       ScanXML
                        scanString: fun {$ S}  {ScanXML S           ''} end
                        scanVS:     fun {$ VS} {ScanXML {VS2S VS}   ''} end
                        scanFile:   fun {$ F}  {ScanXML {LoadFile F} F} end
                        scanURL:    fun {$ U}  {ScanXML {LoadUrl  U} U} end
                       )
      end
   end
end
