%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2001/12/14 23:58:31 $ by $Author: afranke $
%%%   $Revision: 1.3 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% convert a pickled Dtd into syntax rules
%%

functor
   
import
   Application Compiler Error Open OS Pickle Property System

   Boot_Debug at 'x-oz://boot/Debug'
   
   DtdInput   at 'DtdInput.ozf'
   DtdOutput  at 'DtdOutput.ozf'
   Dtd2Parser at 'Dtd2Parser.ozf'
   
require
   IncludeTok(byteString: IncludeTokBS) at 'xml.tok.ozf'
   IncludeSyn(byteString: IncludeSynBS) at 'xml.syn.ozf'
   
prepare

   %% enumeration   
   local
      fun {DoEnumerate Xs Sep F}
         case Xs of nil then ''
         [] X|Xr then Sep#{F X}#{DoEnumerate Xr Sep F} end
      end
   in
      fun {Enumerate Xs Sep F}
         case Xs of nil then ''
         [] X|Xr then {F X}#{DoEnumerate Xr Sep F} end
      end
   end
   local
      fun {Id X} X end
   in
      fun {Enum Xs Sep}
         {Enumerate Xs Sep Id}
      end
   end  

   %% string suffix
   fun {IsSuffix S Suffix}
      {List.isPrefix {Reverse Suffix} {Reverse S}}
   end
   fun {DropSuffix S Suffix}
      if {IsSuffix S Suffix}==false then S
      else
         {Reverse {List.drop {Reverse S} {Length Suffix}}}
      end
   end

define

   AppName = 'dtdtool'
   
   proc {DebugOn}  {Boot_Debug.setRaiseOnBlock {Thread.this} true}  end
   proc {DebugOff} {Boot_Debug.setRaiseOnBlock {Thread.this} false} end
   
   {Property.put 'errors.depth' 100}
   {Property.put 'errors.width' 100}
   
   %%
   %% auxiliuary definitions
   %%
   
   %% file I/O

   proc {LoadVS Url ?S}
      F = {New Open.file init(url:Url)} in
      {F read(list:?S size:all)}
      {F close}
   end
   proc {SaveVS VS FileName}
      F = {New Open.file init(name:FileName flags:[write create truncate])} in
      {F write(vs:VS)}
      {F close}
   end
   proc {CondCopy InputUrl DefaultVS OutputFile}
      VS = if InputUrl == unit
           then DefaultVS
           else {LoadVS InputUrl}
           end
   in   
      {SaveVS VS OutputFile}
   end
   
   %%
   %% get args
   %%
   
   Spec = record(debug(single char:&d type:bool default:false)
                 verbose(single char:&v type:bool default:false)
                 quiet(single char:&q type:bool default:false)
                 input(single char:&i type:string optional:true)
                 output(single char:&o type:string optional:true)
                 
                 mode(single type:atom default:dtd2parser)
                 dtd2parser(single alias:mode#dtd2parser)
                 dtd2pickle(single alias:mode#dtd2pickle)
                 dtd2functor(single alias:mode#dtd2functor)
                 pickle2dtd(single alias:mode#pickle2dtd)
                 pickle2syn(single alias:mode#pickle2syn)
                 pickle2tok(single alias:mode#pickle2tok)
                 pickle2spec(single alias:mode#pickle2spec)
                 pickle2parser(single alias:mode#pickle2parser)
                 
                 dtd(single type:string optional:true)
                 rootelements(multiple type:list(string) default:nil)
                 semantics(single type:string default:false)
                 constructor(single alias:semantics)
                 'include-tokens'(single type:string default:unit)
                 'include-syntaxrules'(single type:string default:unit)
                )
   Args
   try
      Args = {Application.getArgs Spec}
   catch error(ap(usage S)...) then
      {System.showError AppName#": "#S}
      {Application.exit 1}
   end
   
   %% Args.debug

   local
      proc {Skip1 _} skip end
   in
      Debug = {Record.adjoinList dummy
               {Map [show print showInfo printInfo showError printError]
                fun {$ F}
                   F#if Args.debug orelse Args.verbose
                     then System.F
                     else Skip1
                     end
                end}}
   end

   %% Args.input, Args.output

   fun {GetOutput Default}
      {CondSelect Args output Default}
   end
   
   DefaultInput  = '/dev/stdin'
   DefaultOutput = {GetOutput '/dev/stdout'}

   DefaultIOs    = [DefaultInput#DefaultOutput]
   
   Inputs = if {HasFeature Args input} then [Args.input]
            else Args.1 end
   
   if {HasFeature Args output} andthen {Length Inputs}>1 then
      {System.showError AppName#": "#
       "output file specified, but multiple inputs found"}
      {Application.exit 1}
   end
   
   IOs = case Args.mode of dtd2pickle then
            case Inputs of nil then DefaultIOs
            else
               fun {MakeIO X}
                  Base = {DropSuffix X ".dtd"} in
                  (Base#".dtd")#{GetOutput (Base#".dtd.ozp")}
               end
            in
               {Map Inputs MakeIO}
            end
         [] dtd2functor then
            case Inputs of nil then DefaultIOs
            else
               fun {MakeIO X}
                  Base = {DropSuffix X ".dtd"} in
                  (Base#".dtd")#{GetOutput (Base#".dtd.ozf")}
               end
            in
               {Map Inputs MakeIO}
            end
         [] pickle2dtd then
            case Inputs of nil then DefaultIOs
            else
               fun {MakeIO X}
                  Base = {DropSuffix {DropSuffix X ".ozp"} ".dtd"} in
                  (Base#".dtd.ozp")#{GetOutput (Base#".dtd.core")}
               end
            in
               {Map Inputs MakeIO}
            end
         [] pickle2syn then
            case Inputs of nil then DefaultIOs
            else
               fun {MakeIO X}
                  Base = {DropSuffix {DropSuffix X ".ozp"} ".dtd"} in
                  (Base#".dtd.ozp")#{GetOutput (Base#".syn.ozg")}
               end
            in
               {Map Inputs MakeIO}
            end
         [] pickle2parser then
            case Inputs of nil then DefaultIOs
            else
               fun {MakeIO X}
                  Base = {DropSuffix {DropSuffix X ".ozp"} ".dtd"} in
                  (Base#".dtd.ozp")#{GetOutput (Base#".parser.ozf")}
               end
            in
               {Map Inputs MakeIO}
            end
         [] dtd2parser then
            case Inputs of nil then DefaultIOs
            else
               fun {MakeIO X}
                  Base = {DropSuffix X ".dtd"} in
                  (Base#".dtd")#{GetOutput (Base#".parser.ozf")}
               end
            in
               {Map Inputs MakeIO}
            end
         else
            {System.showError AppName#": "#"unknown mode `"#Args.mode#"'"}
            {Application.exit 1}
         end
   
   %%
   %% main
   %%

   fun {GetElements R}
      {Map {Filter R.items fun {$ Kind#_} Kind=='ELEMENT' end}
       fun {$ _#ID} R.elements.ID end}  
   end
   proc {Elements2Parser Elements RootElems SemUrl XmlTokSrc XmlSynSrc
         XmlTokFile XmlSynFile TokFile SynFile SpecFile ParserFile KeepFiles
         ?F}
      %%
      {Debug.printInfo "semantics functor: "}
      if SemUrl==false
      then {Debug.showInfo "none (generic)"}
      else {Debug.showInfo "`"#SemUrl#"'"}
      end
      %%
      {Debug.printInfo "xml tokens       : "}
      if XmlTokSrc==unit
      then {Debug.showInfo "default"}
      else {Debug.showInfo "`"#XmlTokSrc#"'"}
      end
      %%
      {Debug.printInfo "xml syntax rules : "}
      if XmlSynSrc==unit
      then {Debug.showInfo "default"}
      else {Debug.showInfo "`"#XmlSynSrc#"'"}
      end
      %%
      RootElements = 
      case RootElems of nil then
         {Debug.showInfo "Allowing all elements as document root!"}
         {Map Elements fun {$ E} E.name end}
      else   
         RootElems
      end
      {Debug.showInfo "root element(s)  : "#{Enum RootElements ", "}}  
      %%
      {Debug.printInfo "Creating basic tokens file `"#XmlTokFile#"'..."}
      {CondCopy XmlTokSrc IncludeTokBS XmlTokFile}
      {Debug.showInfo " done."}
      %%
      {Debug.printInfo "Creating basic syntax rules file `"#XmlSynFile#"'..."}
      {CondCopy XmlSynSrc IncludeSynBS XmlSynFile}
      {Debug.showInfo " done."}
      %%
      {Debug.printInfo "Creating user tokens file `"#TokFile#"'..."}
      {SaveVS {Dtd2Parser.writeTok Elements} TokFile}
      {Debug.showInfo " done."}
      %%
      {Debug.printInfo "Creating user rules file `"#SynFile#"'..."}
      local
         {DebugOn}
         Y = {Dtd2Parser.writeSyn Elements SemUrl}
         {DebugOff}
      in
         {SaveVS Y SynFile}
      end
      {Debug.showInfo " done."}
      %%
      {Debug.printInfo "Creating elements spec file `"#SpecFile#"'..."} 
      {SaveVS {Dtd2Parser.writeSpec Elements} SpecFile}
      {Debug.showInfo " done."}
      %%
      {Debug.printInfo "Creating gump parser source file `"#ParserFile#"'..."}
      {SaveVS {Dtd2Parser.writeParser SemUrl XmlTokFile XmlSynFile
               TokFile SynFile SpecFile RootElements} ParserFile}
      {Debug.showInfo " done."}
   in
      try
         {Debug.printInfo "Compiling gump parser..."}
         F={Compiler.virtualStringToValue "\\insert '"#ParserFile#"'"}
         {Debug.showInfo " done."}
      catch Err then
         {Error.printException Err}
      finally
         if KeepFiles==false then
            {Debug.printInfo "Removing tmp files..."}
            try
               TmpFiles = [XmlTokFile XmlSynFile TokFile SynFile
                           SpecFile ParserFile]
            in
               {ForAll TmpFiles OS.unlink}
               {Debug.showInfo " done."}
            catch _ then
               {Debug.showInfo " failed."}
            end
         end
         if {IsFree F} then
            {Application.exit 1}
         end
      end
   end  
 
   fun {Elements2ParserTmpFiles Elements SemUrl XmlTok XmlSyn RootElems}
      XmlTokFile = {OS.tmpnam}
      XmlSynFile = {OS.tmpnam}
      TokFile    = {OS.tmpnam}
      SynFile    = {OS.tmpnam}
      SpecFile   = {OS.tmpnam}
      ParserFile = {OS.tmpnam}
   in
      {Elements2Parser Elements RootElems SemUrl XmlTok XmlSyn
       XmlTokFile XmlSynFile TokFile SynFile SpecFile ParserFile false}
   end
   fun {Elements2ParserKeep Elements SemUrl XmlTok XmlSyn RootElems Prefix}
      XmlTokFile = Prefix#".tok-base.ozg"
      XmlSynFile = Prefix#".syn-base.ozg"
      TokFile    = Prefix#".tok.ozg"
      SynFile    = Prefix#".syn.ozg"
      SpecFile   = Prefix#".spec.oz"
      ParserFile = Prefix#".parser.ozg"
   in
      {Elements2Parser Elements RootElems SemUrl XmlTok XmlSyn
       XmlTokFile XmlSynFile TokFile SynFile SpecFile ParserFile true}
   end
   
   fun {GetBase UrlVS}
      UrlS = {VirtualString.toString UrlVS} 
      Last = {Reverse {String.token {Reverse UrlS} &/ $ _}}
   in
      {String.token Last &. $ _}
   end
   fun {GetPrefix I O}
      case O of '/dev/stdout' then
         case I of '/dev/stdin' then
            "./_unknown"
         else "./"#{GetBase I}
         end
      else "./"#{GetBase O}
      end
   end
   
   Main = 
   case Args.mode of dtd2pickle then
      proc {$ I#O}
         {Pickle.save {DtdInput.readDtd I} O}
      end
   [] dtd2functor then
      proc {$ I#O}
         {Pickle.save {DtdInput.toFunctor {DtdInput.readDtd I}} O}
      end
   [] pickle2dtd then
      proc {$ I#O}
         {SaveVS {DtdOutput.writeDtd {Pickle.load I}} O}
      end
   [] pickle2syn then
      proc {$ I#O}
         {SaveVS {Dtd2Parser.writeRules {GetElements {Pickle.load I}}} O}
      end
   [] pickle2parser then
      proc {$ I#O}
         SemUrl    = Args.semantics
         XmlTok    = Args.'include-tokens'
         XmlSyn    = Args.'include-syntaxrules'
         RootElems = Args.rootelements
         Elements  = {GetElements {Pickle.load I}}
      in
         if Args.debug then
            {Pickle.save
             {Elements2ParserKeep Elements SemUrl XmlTok XmlSyn
              RootElems {GetPrefix I O}} O}
         else
            {Pickle.save
             {Elements2ParserTmpFiles Elements SemUrl XmlTok XmlSyn
              RootElems} O}
         end
      end
   [] dtd2parser then
      proc {$ I#O}
         SemUrl    = Args.semantics
         XmlTok    = Args.'include-tokens'
         XmlSyn    = Args.'include-syntaxrules'
         RootElems = Args.rootelements
         Elements  = {GetElements {DtdInput.readDtd I}}
      in
         if Args.debug then
            {Pickle.save
             {Elements2ParserKeep Elements SemUrl XmlTok XmlSyn
              RootElems {GetPrefix I O}} O}
         else
            {Pickle.save
             {Elements2ParserTmpFiles Elements SemUrl XmlTok XmlSyn
              RootElems} O}
         end
      end
   end
   
   {ForAll IOs proc {$ I#O}
                  {Main I#O}
                  if Args.quiet==false andthen O\='/dev/stdout' andthen
                     {HasFeature Args output}==false
                  then
                     {System.showInfo "Wrote "#O}
                  end
               end}
   {Application.exit 0}
end



